import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
import matplotlib
import xarray as xr
from astropy.convolution import CustomKernel
from astropy.convolution import convolve
from astropy.convolution import Gaussian2DKernel
from MITgcmutils import jmd95
from MITgcmutils import rdmds
from MITgcmutils import wrmds
from scipy import signal
from matplotlib.colors import LightSource
import matplotlib.colors as colors
from matplotlib import cm
import xmitgcm
import warnings
from tqdm import tqdm
import oceanspy as ospy
import cmocean
import xgcm


def fmt1(x):
    s = f"{x:.1f}"
    return rf"{s}"

def fmt2(x):
    s = f"{x:.2f}"
    return rf"{s}"

def ds_to_ospy(ds_orig,speed=True,f=True):
    """ds_to_ospy takes an xarray dataset from MITgcm (generated using xmitgcm.open_mdsdataset) and creates an Oceanspy OceanDataset

    Args:
        ds_orig (xr.Dataset): xarray dataset from MITgcm (generated using xmitgcm.open_mdsdataset)
        speed (bool, optional): If false, won't look for velocity information. Defaults to True.

    Returns:
        od (ospy.OceanDataset): Oceanspy OceanDataset
        ds (xr.Dataset): od.ds
        grid (xgcm.Grid): od.grid
    """
    if f:
        # Get the Coriolis parameter field fCoriG (unknown to xmitgcm) by adding it as a coordinate:
        fCoriG = rdmds('../data/fCoriG')
        ds_orig = ds_orig.assign(fCoriG=(('YG', 'XG'), fCoriG))

    # Rename X and Y
    rename_dict = {'XC': 'X', 'YC': 'Y', 'XG': 'Xp1', 'YG': 'Yp1'}
    ds_orig = ds_orig.rename(rename_dict)

    od = ospy.OceanDataset(ds_orig)

    # Set aliases to match OceanSpy naming convention
    od = od.set_aliases(
        {'U': 'UVEL', 'V': 'VVEL', 'W': 'WVEL', 'HFacC': 'hFacC', 'HFacW': 'hFacW', 'HFacS': 'hFacS', 'Temp': 'THETA',
         'S': 'SALT'}, overwrite=True)

    # Create co-ordinate system (match to output of xgcm.Grid(ds)) see https://xgcm.readthedocs.io/en/latest/grids.html
    od = od.manipulate_coords(fillna=False, coords1Dfrom2D=False, coords2Dfrom1D=True, coordsUVfromG=False)
    od = od.set_grid_coords(
        {
            "time": {"time": 0},
            'Y': {'Y': None, 'Yp1': -0.5},
            'X': {'X': None, 'Xp1': -0.5},
            'Z': {'Z': None, 'Zp1': -0.5, 'Zu': 0.5, 'Zl': -0.5},
        },
        overwrite=True)
    grid = od._grid
    #od = od.compute.Brunt_Vaisala_frequency()
    if speed:
        # get flow speed:
        speed = np.sqrt(grid.interp(od.dataset.UVEL, 'X', boundary='extend') ** 2 + grid.interp(od.dataset.VVEL, 'Y',
                                                                                                boundary='extend') ** 2)
        speed = speed.rename('speed')
        od = od.merge_into_oceandataset(speed)
    ds = od.dataset

    return od, ds, grid


def smooth_2D(field,dx,dy,width=20000):
    """smooth_2D performs a uniform convolution (equivalent to a 2D moving average) of some 2D field

    Args:
        field (np.ndarray): 2D input field, dimensions (y,x)
        dx (float): grid spacing in x (m)
        dy (float): grid spacing in y (m)
        width (int, optional): dimensions of (width x width) box to average over (m). Defaults to 20000.

    Returns:
        np.ndarray: 2D smoothed field, dimensions (y,x)
    """

    #smooth_2D performs a uniform convolution (equivalent to a 2D moving average) of some 2D field

    # :field: 2D input field, dimensions (y,x) dimensions
    # :dx: grid spacing in x (in same units as width)
    # :dy: grid spacing in y (in same units as width)
    # :width: dimensions of (width x width) box to average over 
    # :return: smoothed 2D field with same dimensions as 'field'
    
    Nx = 2*int(width/dx/2) + 1
    Ny = 2*int(width/dy/2) + 1
    
    return convolve(field,np.ones((Ny,Nx))/Nx/Ny,boundary='extend')




def find_in_HAB(ds_grid, field3D, HAB):
    """find_in_HAB finds a 2D field from a 3D field and a given height above bottom.

    Args:
        ds_grid (xr.Dataset): dataset containing information about the model grid
        field3D (np.ndarray): some 3D flow field
        HAB (float):Height above bottom (m) at which to find the flow

    Returns:
        field2D (np.ndarray): 2D lat-lon field containing data from 3D field at HAB.
    """
    nz = ds_grid.Z.values.shape[0]
    nx = ds_grid.XC.values.shape[0]
    ny = ds_grid.YC.values.shape[0]
    Zmesh = np.meshgrid( ds_grid.YC.values, ds_grid.Z.values,ds_grid.XC.values)[1]
    depth_of_strike_mesh = np.tile(np.expand_dims(-ds_grid.Depth.values + HAB,0),[nz,1,1])
    iZ_of_strike = np.argmin(np.abs(depth_of_strike_mesh-Zmesh), axis=0)
    
    
    field2D = np.zeros((ny,nx))
    for ix in tqdm(range(nx)):
        for iy in range(ny):
            if iZ_of_strike[iy,ix] > 55:
                field2D[iy,ix] = field3D[iZ_of_strike[iy,ix],iy,ix]
            else:
                field2D[iy,ix] = np.nan
            
    return field2D

def find_in_HAB_av(ds_grid,field3D,HAB_lower,HAB_upper):
    """find_in_HAB_av finds a 2D field from a 3D field and a two height above bottom limits to average between.

    Args:
        ds_grid (xr.Dataset): dataset containing information about the model grid
        field3D (np.ndarray): some 3D flow field
        HAB_lower (float):Lower limit of height above bottom (m) 
        HAB_upper (float):Upper limit of height above bottom (m) 

    Returns:
        field2D (np.ndarray): 2D lat-lon field containing data from 3D field averaged between HAB limits.
    """
    nz = ds_grid.Z.values.shape[0]
    nx = ds_grid.XC.values.shape[0]
    ny = ds_grid.YC.values.shape[0]
    Zmesh = np.meshgrid( ds_grid.YC.values, ds_grid.Z.values,ds_grid.XC.values)[1]
    depth_of_strike_mesh_lower = np.tile(np.expand_dims(-ds_grid.Depth.values + HAB_lower,0),[nz,1,1])
    depth_of_strike_mesh_upper = np.tile(np.expand_dims(-ds_grid.Depth.values + HAB_upper,0),[nz,1,1])
    iZ_of_strike_lower = np.argmin(np.abs(depth_of_strike_mesh_lower-Zmesh), axis=0)
    iZ_of_strike_upper = np.argmin(np.abs(depth_of_strike_mesh_upper-Zmesh), axis=0)
    
    
    field2D = np.zeros((ny,nx))
    for ix in tqdm(range(nx)):
        for iy in range(ny):
            if iZ_of_strike_upper[iy,ix] > 55:
                field2D[iy,ix] = np.nanmean(field3D[iZ_of_strike_upper[iy,ix]:iZ_of_strike_lower[iy,ix]+1,iy,ix],axis=0)
            else:
                field2D[iy,ix] = np.nan
            
    return field2D

def do_WMT(ds):
    # Get the Coriolis parameter field fCoriG (unknown to xmitgcm) by adding it as a coordinate:
    fCoriG = rdmds(
            '../data/fCoriG')
    ds = ds.assign(fCoriG = (('YG','XG'),fCoriG))

    # Rename X and Y
    rename_dict = {'XC': 'X', 'YC': 'Y', 'XG': 'Xp1', 'YG': 'Yp1'}
    ds = ds.rename(rename_dict)
    od = ospy.OceanDataset(ds)

    # Set aliases to match OceanSpy naming convention
    od = od.set_aliases({'U':'UVEL','V':'VVEL','W':'WVEL','HFacC':'hFacC','HFacW':'hFacW','HFacS':'hFacS','Temp':'THETA','S':'SALT'},overwrite = True)

    # Create co-ordinate system (match to output of xgcm.Grid(ds)) see https://xgcm.readthedocs.io/en/latest/grids.html
    od = od.manipulate_coords(fillna = False, coords1Dfrom2D=False, coords2Dfrom1D=True, coordsUVfromG=False)
    od = od.set_grid_coords(
           {
            "time": {"time": 0},
            'Y': {'Y': None, 'Yp1': -0.5},
            'X': {'X': None, 'Xp1': -0.5},
            'Z': {'Z': None, 'Zp1': -0.5, 'Zu': 0.5, 'Zl': -0.5},
            },
            overwrite=True )
    od = od.set_grid_periodic(['Y','X']) # We don't care about boundary values in sponge layer - this removes errors
    grid = od._grid

    # Compute N2
    #od = od.compute.Brunt_Vaisala_frequency()

    # Define buoyancy flux and add to dataset
    Bflux = -od.dataset.KPPdiffT*od.dataset.N2

    # N2 is on lower face of cell and is undefined on topography (since sigma is a nan in centre cell below).
    # Set Bflux = 0 on Zl coord at topography (by setting to zero everywhere it is nan), can ignore hfac issues since hfac = 0 or 1 in this simulation

    Bflux = xr.where(np.isnan(Bflux),0,Bflux)
    Bflux = Bflux.rename('Bflux')
    od = od.merge_into_oceandataset(Bflux)

    # Calculate N^2 centre points
    N2_on_Z = grid.interp(od.dataset.N2.ffill('Zl',1),'Z',boundary='extend')
    N2_on_Z = N2_on_Z.rename('N2_on_Z')
    od = od.merge_into_oceandataset(N2_on_Z)

    # Compute gradient of buoyancy flux, this is on centre points.
    od = od.compute.gradient(varNameList="Bflux",axesList="Z")

    # Define a volume element
    dV = od.dataset.drF*od.dataset.rA*od.dataset.hFacC

    # Define density levels below which to calculate integral
    ng = 51
    gamma_bounds = np.linspace(28,28.3,ng)

    HABC = (od.dataset.Z + od.dataset.Depth).load()
    ds= od.dataset

    # Take dBflux_from_mean_dZ - we are calculating this from the mean quantities.
    weighted_dBfluxdz = dV*ds.dBflux_dZ

    # Topography mask
    weighted_dBfluxdz = xr.where(od.dataset.maskC,weighted_dBfluxdz,np.nan).load()
    
    # Make a mask for no bottom gridcell:
    no_bottom_mask = xr.where(np.isnan(xr.where(ds.maskC==1,np.nan,0).bfill('Z',1)),1,0)
    weighted_dBfluxdz_no_bottom = xr.where(no_bottom_mask==1,weighted_dBfluxdz,np.nan).load()

    # Up and down components
    weighted_dBfluxdz_up = xr.where(weighted_dBfluxdz < 0,weighted_dBfluxdz,np.nan)
    weighted_dBfluxdz_down = xr.where(weighted_dBfluxdz > 0,weighted_dBfluxdz,np.nan)
    
    weighted_dBfluxdz_up_no_bottom = xr.where(weighted_dBfluxdz_no_bottom < 0,weighted_dBfluxdz_no_bottom,np.nan)
    weighted_dBfluxdz_down_no_bottom = xr.where(weighted_dBfluxdz_no_bottom > 0,weighted_dBfluxdz_no_bottom,np.nan)

    # Set up arrays for integral
    Q_bdy = np.zeros((ng))
    Q_up_bdy = np.zeros((ng))
    Q_down_bdy = np.zeros((ng))

    Q_int = np.zeros((ng))
    Q_up_int = np.zeros((ng))
    Q_down_int = np.zeros((ng))

    Q_stable = np.zeros((ng))
    Q_up_stable = np.zeros((ng))
    Q_down_stable = np.zeros((ng))

    Q_unstable = np.zeros((ng))
    Q_up_unstable = np.zeros((ng))
    Q_down_unstable = np.zeros((ng))
    Q_total = np.zeros((ng))
    
    Q_bdy_no_bottom = np.zeros((ng))
    Q_up_bdy_no_bottom = np.zeros((ng))
    Q_down_bdy_no_bottom = np.zeros((ng))

    Q_int_no_bottom = np.zeros((ng))
    Q_up_int_no_bottom = np.zeros((ng))
    Q_down_int_no_bottom = np.zeros((ng))

    Q_stable_no_bottom = np.zeros((ng))
    Q_up_stable_no_bottom = np.zeros((ng))
    Q_down_stable_no_bottom = np.zeros((ng))

    Q_unstable_no_bottom = np.zeros((ng))
    Q_up_unstable_no_bottom = np.zeros((ng))
    Q_down_unstable_no_bottom = np.zeros((ng))
    Q_total_no_bottom = np.zeros((ng))

    Sigma0 = ds.Sigma0.load()
    N2_on_Z = ds.N2_on_Z.load()

    for ig in tqdm(range(0,ng)):


        Q_bdy[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz,np.nan).sum()
        Q_up_bdy[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz_up,np.nan).sum()
        Q_down_bdy[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz_down,np.nan).sum()

        Q_int[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz,np.nan).sum()
        Q_up_int[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz_up,np.nan).sum()
        Q_down_int[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz_down,np.nan).sum()
        
        Q_bdy_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz_no_bottom,np.nan).sum()
        Q_up_bdy_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz_up_no_bottom,np.nan).sum()
        Q_down_bdy_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC < 250),weighted_dBfluxdz_down_no_bottom,np.nan).sum()

        Q_int_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz_no_bottom,np.nan).sum()
        Q_up_int_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz_up_no_bottom,np.nan).sum()
        Q_down_int_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (HABC >= 250),weighted_dBfluxdz_down_no_bottom,np.nan).sum()

        Q_stable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz,np.nan).sum()
        Q_up_stable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz_up,np.nan).sum()
        Q_down_stable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz_down,np.nan).sum()

        Q_unstable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz,np.nan).sum()
        Q_up_unstable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz_up,np.nan).sum()
        Q_down_unstable[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz_down,np.nan).sum()
        
        Q_stable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz_no_bottom,np.nan).sum()
        Q_up_stable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz_up_no_bottom,np.nan).sum()
        Q_down_stable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z > 0),weighted_dBfluxdz_down_no_bottom,np.nan).sum()

        Q_unstable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz_no_bottom,np.nan).sum()
        Q_up_unstable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz_up_no_bottom,np.nan).sum()
        Q_down_unstable_no_bottom[ig] = xr.where((Sigma0 > gamma_bounds[ig]) & (N2_on_Z <= 0) & (~np.isnan(N2_on_Z)),weighted_dBfluxdz_down_no_bottom,np.nan).sum()

        Q_total[ig] = xr.where((ds.Sigma0 > gamma_bounds[ig]),weighted_dBfluxdz,np.nan).sum()
        Q_total_no_bottom[ig] = xr.where((ds.Sigma0 > gamma_bounds[ig]),weighted_dBfluxdz_no_bottom,np.nan).sum()
        
        
    WMT_up_bdy = np.diff(Q_up_bdy)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_bdy = np.diff(Q_down_bdy)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_int = np.diff(Q_up_int)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_int = np.diff(Q_down_int)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_stable = np.diff(Q_up_stable)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_stable = np.diff(Q_down_stable)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_unstable = np.diff(Q_up_unstable)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_unstable = np.diff(Q_down_unstable)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_total = np.diff(Q_total)/np.diff(gamma_bounds)*1027/9.81/1e6
    
    WMT_up_bdy_no_bottom = np.diff(Q_up_bdy_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_bdy_no_bottom = np.diff(Q_down_bdy_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_int_no_bottom = np.diff(Q_up_int_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_int_no_bottom = np.diff(Q_down_int_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_stable_no_bottom = np.diff(Q_up_stable_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_stable_no_bottom = np.diff(Q_down_stable_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_up_unstable_no_bottom = np.diff(Q_up_unstable_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6
    WMT_down_unstable_no_bottom = np.diff(Q_down_unstable_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6

    WMT_total_no_bottom = np.diff(Q_total_no_bottom)/np.diff(gamma_bounds)*1027/9.81/1e6

    gamma_mid=0.5*(gamma_bounds[0:-1]+gamma_bounds[1:])
    
    ds_out = xr.Dataset(
        data_vars=dict(


            WMT_up_bdy=(["gamma_mid"], WMT_up_bdy),
            WMT_down_bdy=(["gamma_mid"], WMT_down_bdy),
            WMT_up_int=(["gamma_mid"], WMT_up_int),
            WMT_down_int=(["gamma_mid"], WMT_down_int),
            WMT_up_stable=(["gamma_mid"], WMT_up_stable),
            WMT_down_stable=(["gamma_mid"], WMT_down_stable),
            WMT_up_unstable=(["gamma_mid"], WMT_up_unstable),
            WMT_down_unstable=(["gamma_mid"], WMT_down_unstable),
            WMT_total=(["gamma_mid"], WMT_total),
            WMT_up_bdy_no_bottom=(["gamma_mid"], WMT_up_bdy_no_bottom),
            WMT_down_bdy_no_bottom=(["gamma_mid"], WMT_down_bdy_no_bottom),
            WMT_up_int_no_bottom=(["gamma_mid"], WMT_up_int_no_bottom),
            WMT_down_int_no_bottom=(["gamma_mid"], WMT_down_int_no_bottom),
            WMT_up_stable_no_bottom=(["gamma_mid"], WMT_up_stable_no_bottom),
            WMT_down_stable_no_bottom=(["gamma_mid"], WMT_down_stable_no_bottom),
            WMT_up_unstable_no_bottom=(["gamma_mid"], WMT_up_unstable_no_bottom),
            WMT_down_unstable_no_bottom=(["gamma_mid"], WMT_down_unstable_no_bottom),
            WMT_total_no_bottom=(["gamma_mid"], WMT_total_no_bottom),

        ),
        coords=dict(
            gamma_mid=(["gamma_mid"], gamma_mid),
            gamma_bounds=(["gamma_bounds"], gamma_bounds),

        ),
        attrs=dict(description="Water mass transformation from L2 KPP")
        )

    ds_out.WMT_up_bdy.attrs["long_name"] = "Positive water mass transformation (< 250 m above bottom)"    
    ds_out.WMT_up_bdy.attrs["units"] = "Sv"
    ds_out.WMT_up_int.attrs["long_name"] = "Positive water mass transformation (> 250 m above bottom)"    
    ds_out.WMT_up_int.attrs["units"] = "Sv"
    ds_out.WMT_down_bdy.attrs["long_name"] = "Negative water mass transformation (< 250 m above bottom)"    
    ds_out.WMT_down_bdy.attrs["units"] = "Sv"
    ds_out.WMT_down_int.attrs["long_name"] = "Negative water mass transformation (> 250 m above bottom)"    
    ds_out.WMT_down_int.attrs["units"] = "Sv"

    ds_out.WMT_up_stable.attrs["long_name"] = "Positive water mass transformation (N^2 > 0)"    
    ds_out.WMT_up_stable.attrs["units"] = "Sv"
    ds_out.WMT_up_unstable.attrs["long_name"] = "Positive water mass transformation (N^2 < 0)"    
    ds_out.WMT_up_unstable.attrs["units"] = "Sv"
    ds_out.WMT_down_stable.attrs["long_name"] = "Negative water mass transformation (N^2 > 0)"    
    ds_out.WMT_down_stable.attrs["units"] = "Sv"
    ds_out.WMT_down_unstable.attrs["long_name"] = "Negative water mass transformation (N^2 < 0)"    
    ds_out.WMT_down_unstable.attrs["units"] = "Sv"

    ds_out.WMT_total.attrs["long_name"] = "Water mass transformation"    
    ds_out.WMT_total.attrs["units"] = "Sv"
    
    ds_out.WMT_up_bdy_no_bottom.attrs["long_name"] = "Positive water mass transformation (< 250 m above bottom, no bottom cell)"    
    ds_out.WMT_up_bdy_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_up_int_no_bottom.attrs["long_name"] = "Positive water mass transformation (> 250 m above bottom, no bottom cell)"    
    ds_out.WMT_up_int_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_down_bdy_no_bottom.attrs["long_name"] = "Negative water mass transformation (< 250 m above bottom, no bottom cell)"    
    ds_out.WMT_down_bdy_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_down_int_no_bottom.attrs["long_name"] = "Negative water mass transformation (> 250 m above bottom, no bottom cell)"    
    ds_out.WMT_down_int_no_bottom.attrs["units"] = "Sv"

    ds_out.WMT_up_stable_no_bottom.attrs["long_name"] = "Positive water mass transformation (N^2 > 0, no bottom cell)"    
    ds_out.WMT_up_stable_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_up_unstable_no_bottom.attrs["long_name"] = "Positive water mass transformation (N^2 < 0, no bottom cell)"    
    ds_out.WMT_up_unstable_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_down_stable_no_bottom.attrs["long_name"] = "Negative water mass transformation (N^2 > 0, no bottom cell)"    
    ds_out.WMT_down_stable_no_bottom.attrs["units"] = "Sv"
    ds_out.WMT_down_unstable_no_bottom.attrs["long_name"] = "Negative water mass transformation (N^2 < 0, no bottom cell)"    
    ds_out.WMT_down_unstable_no_bottom.attrs["units"] = "Sv"

    ds_out.WMT_total_no_bottom.attrs["long_name"] = "Water mass transformation, no bottom cell"    
    ds_out.WMT_total_no_bottom.attrs["units"] = "Sv"

    ds_out.gamma_mid["long_name"] = "Neutral density (interval midpoint)"
    ds_out.gamma_mid["units"] = "kg/m^3"
    ds_out.gamma_bounds["long_name"] = "Neutral density (interval endpoint)"
    ds_out.gamma_bounds["units"] ="kg/m^3"

    return ds_out

def average_hab(od,field):
        """
        This function takes a field with dimensions (Z,Y,X) and finds it's average over the domain in height above bottom (HAB) coordinates.
        Input whole ocean data set 'od', 'field' to find in HAB coords at centre coordinates, resolution of resulting vector 'nz'. 
        """
        grid = od.grid
        field = np.squeeze(field)
        # Fill in nans
        field = xr.where(field == 0,np.nan, field)
        # Make sure all input fields are on same grid
        if 'Zl' in field.dims:
            # First fill in the nans so we don't lose the bottom cell (Bflux is already set to zero on the topography so won't be changed)
            field = field.ffill(dim = 'Zl')

            # Now interpolate onto Z grid
            field = grid.interp(field,'Z',boundary = 'extend')

        # Ensure topography is masked
        field = xr.where(od.dataset.maskC,field,np.nan)

        # Define a weighted area element and vertical grid length 
        dA = od.dataset.rA*od.dataset.hFacC
        dA = xr.where(od.dataset.maskC,dA,np.nan)
        
        # Get rid of infinities
        field = xr.where(np.isinf(field),np.nan,field)
        
        # Weight the field with area elements
        weighted_field = np.squeeze(field*dA)
        
        # Define the coordinate z vector for height above bottom, with dz = 25 since most of domain has this step
        zC_AB = np.linspace(0,2000,81) + 12.5
        
        # Interpolate onto an evenly spaced grid
        z_even = np.arange(-5,-5500,-25)
        weighted_field_interp = weighted_field.interp(Z = z_even,method="nearest")
        dA_interp = dA.interp(Z = z_even,method="nearest")
        
        # Fill in any weird surface nans so they are counted in the mask
        weighted_field_interp = xr.where(weighted_field_interp.Z > -100,0,weighted_field_interp)
        dA_interp = xr.where(dA_interp.Z > -100,0,dA_interp)

        # Find indices of bottom cell in new grid
        mask = xr.where(~np.isnan(weighted_field_interp), 1, 0)
        bottom_ind = mask.sum(axis = 0).values -1
        
        # Take away surface stuff - above 1000m
        weighted_field_interp = xr.where((weighted_field_interp.Z) < -1000,weighted_field_interp,np.nan)
        dA_interp = xr.where((weighted_field_interp.Z) < -1000,dA_interp,np.nan)
        
        field_load = np.squeeze(weighted_field_interp.values)
        weights_load = np.squeeze(dA_interp.values)
        
        # Now some lots of nans at the top of the array to avoid the index reaching for values outside of the field
        field_load_cat = np.concatenate((np.full((81,400,1480),np.nan),field_load),axis = 0)
        weights_load_cat = np.concatenate((np.full((81,400,1480),np.nan),weights_load),axis = 0)     
        
        # Initialise HAB 3D field
        profile_AB = np.zeros_like(zC_AB)
        
        for iz in tqdm(range(81)):
            m,n = bottom_ind.shape
            I,J = np.ogrid[:m,:n]
            out = field_load_cat[81+bottom_ind  -iz, I, J]
            weights_out = weights_load_cat[81+bottom_ind -iz, I, J]
            profile_AB[iz] = np.nansum(out)/np.nansum(weights_out)

            
        
        return zC_AB,profile_AB